
-- PARALLEL LINE FUNCTION ************************************************
-- p1 and p2 are 2D line endpoints
-- offset is distance between original line and parallel line
-- side is boolean value, true => right, false => left
-- Returned values: q1 and q2 are offset line endpoints
-- normalize (q2-q1) is unit vector along offset line
global AcadFloat1
global MlineDraw

fn ParLine p1: p2: offset: side: =
(
	-- VARIABLE DECLARATIONS
	local od,z_dir,q1,q2
	
	-- TEST FOR SIDE
	if side == #right
	then (z_dir = [0,0,-1])
	else (z_dir = [0,0,1])
	
	-- PERFORM CROSS PRODUCT TO DETERMINE OFFSET DIRECTION, (od)
	od = normalize(cross z_dir (p2-p1))
	
	-- PARALLEL LINE ENDPOINTS
	q1 = p1 + od*offset
	q2 = p2 + od*offset
	
	--ENSURE Z COMPONENT IS ZERO
	q1.z = 0
	q2.z = 0
	
	-- RETURN ARRAY OF VALUES
	#(q1,q2,normalize(q2-q1))
)

-- 2D MLINE INTERSECTION FUNCTION **********************************************
-- This function is NOT general case, specific to mline application
-- assumption parallel=> colinear & p2 = q1
-- determines intersection point of lines formed by supplied points
-- supplied points are type point2 
fn MLineIntersect p1 p2 q1 q2 =
(
	-- VARIABLE DECLARATIONS
	local A,B,C,r,s,Denom
	
	-- SETUP FOR INTERSECTION POINT CALCULATION
	A = p2-p1
	B = q1-q2
	C = q1-p1
	
	-- DENOMINATOR NEAR 0 IMPLIES LINES APPROX PARALLEL
	if abs(Denom = A.y*B.x-B.y*A.x) < 0.0001 then
	(	
		[(p2.x+q1.x)/2.0,(p2.y+q1.y)/2.0,0.0]
	)	
	else
	(	
		-- PARAMETRIC CONSTANTS FROM CRAMER'S RULE
		r = (C.y*B.x-B.y*C.x) / Denom	-- associated with line p1p2
		-- s = (A.y*C.x-C.y*A.x) / Denom	-- associated with line q1q2
	
		-- DETERMINE INTERSECTION POINT
		p = p1 + r * A
		[p.x,p.y,0.0]
	)
)

-- ROLLOUT DEFINITION
rollout roMline "Mline Settings"
(
	-- DECLARATIONS
	local btn_size = 25		
	
	-- USER INTERFACE ITEMS
	radiobuttons MlineJust "Justification" \
	labels:#("Top", "Zero", "Bottom") \
	default:1 align:#left offset:[0,0] columns:1
	
	spinner MlineScale  "Scale" type:#float \ 
	range:[0.0,10000,8.0]  fieldwidth:45 scale:0.1 \
	align:#left 
 

	bitmap bm_mline_setup  fileName:"mline.bmp" \ 
	align:#right offset:[-10,-90]
	
	button btnOkGo "Continue" across:2 align:#right offset:[-10,0]
	button btnCancelGo "Cancel" align:#left offset:[10,0]
	
	-- EVENT HANDLERS
	on btnOkGo pressed do
	(
		closeRolloutFloater AcadFloat1
		MlineDraw MlineJust.state MlineScale.value
	)
	
	on btnCancelGo pressed do closeRolloutFloater AcadFloat1
	
)

-- MLINE DRAWOUT ROUTINE ********************************************************************************
fn MlineDraw Just dist =
(
	-- VARIABLE DECLARATIONS
	local SpLH,SpRH,SpCL
	local CL1,CL2,CL3,CL4
	local RH1,RH2,RH3
	local LH1,LH2,LH3
	local aR=bR=cR=dR=fR=gR=[0,0,0]	
	local aL=bL=cL=dL=eL=fL=[0,0,0]
	local CarryOn = true
	local TolVal = 0.0005
	local WorkCS,WorkVP,WorkTM,VT
	local TForm = false
	local LhOffset,RhOffset
	local prompts = #(),LPrompt
	local j,k,UserGrid
	
	-- MLINE MOUSE CALLBACK ROUTINE **********************************************************************
	fn MlineHdlr p2 spln:MlineDraw.SpCL =
	(
		-- DETERMINE NUMBER OF KNOTS IN SPLINESHAPE
		nK = numKnots spln 1
		
		if (length(p2 - (getKnotPoint spln 1 nK)) > 0.0001) do 		
		(
	
			if nK == 2 then	-- THIS WILL BE THE CASE THE FIRST TIME THROUGH MOUSE HANDLER
			(
				-- GET CENTERLINE POINTS
				MlineDraw.CL1 = getknotpoint spln 1 1
				MlineDraw.CL2 = p2
				
				-- SET CURRENT KNOT POINT
				setKnotPoint spln 1 2 MlineDraw.CL2
				
				-- TRANSFORM POINTS IF REQUIRED
				if MlineDraw.TForm do
				(
					MlineDraw.CL1 = PointConditioner testpt:(MlineDraw.CL1*MlineDraw.WorkTM) tolerance:MlineDraw.TolVal
					MlineDraw.CL2 = PointConditioner testpt:(MlineDraw.CL2*MlineDraw.WorkTM) tolerance:MlineDraw.TolVal
				)
				
				-- GET POINTS DEFINING PARALLEL LINES TO THE RIGHT AND LEFT
				local rhRes = ParLine p1:MlineDraw.CL1 p2:MlineDraw.CL2 offset:MlineDraw.RhOffset side:#right
				local lhRes = ParLine p1:MlineDraw.CL1 p2:MlineDraw.CL2 offset:MlineDraw.LhOffset side:#left
														
				-- FIND INTERSECTION POINTS ON RH AND LH SIDE
				MlineDraw.RH1 = rhRes[1]	-- line startpt 
				MlineDraw.RH2 = rhRes[2]	-- line endpoint
				MlineDraw.LH1= lhRes[1]	-- line startpt 
				MlineDraw.LH2= lhRes[2]	-- line endpoint
				
				-- TRANSFORM POINTS BACK IF REQUIRED
				if MlineDraw.TForm do
				(
					MlineDraw.RH1 = PointConditioner testpt:(MlineDraw.RH1*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
					MlineDraw.RH2 = PointConditioner testpt:(MlineDraw.RH2*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
					MlineDraw.LH1 = PointConditioner testpt:(MlineDraw.LH1*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
					MlineDraw.LH2 = PointConditioner testpt:(MlineDraw.LH2*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
		 		)
				
				-- SET RIGHT AND LEFT KNOT POINTS
				setKnotPoint MlineDraw.SpRH 1 (nK-1) MlineDraw.RH1
				setKnotPoint MlineDraw.SpRH 1 nK MlineDraw.RH2
				setKnotPoint MlineDraw.SpLH 1 (nK-1) MlineDraw.LH1
				setKnotPoint MlineDraw.SpLH 1 nK MlineDraw.LH2
				
				-- UPDATE SHAPE CACHE AND REDRAW
				updateshape spln
				updateshape MlineDraw.SpRH
				updateshape MlineDraw.SpLH
				redrawviews()
			) 
			else -- THIS WILL BE THE CASE FOR ALL OTHER PASSES THROUGH THE MOUSE HANDLER
			(
				-- GET CENTERLINE POINTS
				MlineDraw.CL1 = getKnotPoint spln 1 (nK-2)	
				MlineDraw.CL2 = getKnotPoint spln 1 (nK-1)	
				MlineDraw.CL3 = p2		
						
				-- SET FINAL KNOT ON CENTERLINE
				setKnotPoint spln 1 nK MlineDraw.CL3
				
				
				-- TRANSFORM POINTS IF REQUIRED
				if MlineDraw.TForm do
				(
					MlineDraw.CL1 = PointConditioner testpt:(MlineDraw.CL1*MlineDraw.WorkTM) tolerance:MlineDraw.TolVal
					MlineDraw.CL2 = PointConditioner testpt:(MlineDraw.CL2*MlineDraw.WorkTM) tolerance:MlineDraw.TolVal
					MlineDraw.CL3 = PointConditioner testpt:(MlineDraw.CL3*MlineDraw.WorkTM) tolerance:MlineDraw.TolVal
				)
		
				-- GET POINTS DEFINING PARALLEL LINES TO THE RIGHT
				rhRes1 = ParLine p1:MlineDraw.CL1 p2:MlineDraw.CL2 offset:MlineDraw.RhOffset side:#right
				rhRes2 = ParLine p1:MlineDraw.CL2 p2:MlineDraw.CL3 offset:MlineDraw.RhOffset side:#right
								
														
				-- FIND INTERSECTION POINTS ON RH SIDE
				MlineDraw.aR = rhRes1[1]	-- 1st line startpt 
				MlineDraw.bR = rhRes1[2]	-- 1st line endpoint
				MlineDraw.cR = rhRes2[1]	-- 2nd line startpt
				MlineDraw.dR = rhRes2[2]	-- 2nd line endpoint
									
				MlineDraw.RH1 = MLineIntersect [MlineDraw.aR.x,MlineDraw.aR.y] [MlineDraw.bR.x,MlineDraw.bR.y] [MlineDraw.cR.x,MlineDraw.cR.y] [MlineDraw.dR.x,MlineDraw.dR.y]
				MlineDraw.RH2 = MlineDraw.dR
			
				-- GET POINTS DEFINING PARALLEL LINES TO THE LEFT
				lhRes1 = ParLine p1:MlineDraw.CL1 p2:MlineDraw.CL2 offset:MlineDraw.LhOffset side:#left
				lhRes2 = ParLine p1:MlineDraw.CL2 p2:MlineDraw.CL3 offset:MlineDraw.LhOffset side:#left
	
				-- FIND INTERSECTION POINTS ON LH SIDE
				MlineDraw.aL = lhRes1[1]	-- 1st line startpt 
				MlineDraw.bL = lhRes1[2]	-- 1st line endpoint
				MlineDraw.cL = lhRes2[1]	-- 2nd line startpt
				MlineDraw.dL = lhRes2[2]	-- 2nd line endpoint
									
				MlineDraw.LH1 = MLineIntersect [MlineDraw.aL.x,MlineDraw.aL.y] [MlineDraw.bL.x,MlineDraw.bL.y] [MlineDraw.cL.x,MlineDraw.cL.y] [MlineDraw.dL.x,MlineDraw.dL.y]
				MlineDraw.LH2 = MlineDraw.dL
	
				-- TRANSFORM POINTS BACK IF REQUIRED
				if MlineDraw.TForm do
				(
					MlineDraw.RH1 = PointConditioner testpt:(MlineDraw.RH1*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
					MlineDraw.RH2 = PointConditioner testpt:(MlineDraw.RH2*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
					MlineDraw.LH1 = PointConditioner testpt:(MlineDraw.LH1*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
					MlineDraw.LH2 = PointConditioner testpt:(MlineDraw.LH2*inverse(MlineDraw.WorkTM)) tolerance:MlineDraw.TolVal
	 			)
				
				-- SET RIGHT AND LEFT KNOT POINTS
				setKnotPoint MlineDraw.SpRH 1 (nK-1) MlineDraw.RH1
				setKnotPoint MlineDraw.SpRH 1 nK MlineDraw.RH2
				setKnotPoint MlineDraw.SpLH 1 (nK-1) MlineDraw.LH1
				setKnotPoint MlineDraw.SpLH 1 nK MlineDraw.LH2
				
				-- UPDATE SHAPE CACHE AND REDRAW
				updateshape spln
				updateshape MlineDraw.SpRH
				updateshape MlineDraw.SpLH
				redrawviews()
			)
		)
	)			
	
	-- CODE BODY ******************************************************************************************
	-- ****************************************************************************************************
	
	-- SET SYSTEM PARAMETERS
	clearlistener()
	EscapeEnable = false
	CS = set coordsys world
		
	-- SET OFFSET
	LhOffset = RhOffset = dist/2.0
	
	-- INITIALIZE PROMPTS
	prompts[1] = "Start point: \n"
	prompts[2] = "To point: \n"
	prompts[3] = "Undo/<To point>: \n"
	prompts[4] = "Close/Undo/<To point>: \n"
	prompts[5] = "OK\n"
	
	-- GET INITIAL MLINE POINT
	format prompts[1]
	OriginPt = p1 = pickpoint()
	
	if (ClassOf p1 != Point3) do
	(
		EscapeEnable = true
		setListenerSel #(-1,-1)
		return (format "*Cancel*\nOK\n")
	)
	format "\n"
	
	-- DETERMINE WHETHER USER GRID EXISTS AND IS ACTIVE
	j = getCPTM()	
	k = matrix3 1
	if (j.row1==k.row1 and j.row2==k.row2 and j.row3==k.row3 and j.row4==k.row4) 
	then UserGrid = false
	else UserGrid = true
	
	-- DETERMINE WHAT TRANSFORMATION IS REQD (IF NOT REQD SET TO WCS; MATRIX 3 1)
	if UserGrid 
	then 
	( 
		WorkTM = inverse j
		TForm = true
	)
	else 
	(
		case view_type() of 
		(
			#perspective: 	( TForm = false ; WorkTM = matrix3 1 )
			#user:			( TForm = false ; WorkTM = matrix3 1 )
			#top:			( Tform = false ; WorkTM = matrix3 1 )
			default:		( TForm = true ; WorkTM = getViewTM())
		)
	)

	-- SAVE VIEW TYPE FOR CHECKING USER INPUT
	WorkVP = view_type()
	
	-- INITIALIZE SPLINESHAPES
	-- CENTER LINE
	SpCL = splineshape pos:[0,0,0]
	if TForm do	SpCL.transform = inverse(WorkTM)
	addnewspline SpCL
	addknot SpCL 1 #corner #line OriginPt
	addknot SpCL 1 #corner #line [0,0,0]

	-- RIGHT HAND SIDE
	SpRH = splineshape pos:[0,0,0]
	if TForm do	SpRH.transform = inverse(WorkTM)
	SpRH.wirecolor = SpCL.wirecolor
	addnewspline SpRH
	addknot SpRH 1 #corner #line [0,0,0]
	addknot SpRH 1 #corner #line [0,0,0]

	-- LEFT HAND SIDE
	SpLH = splineshape pos:[0,0,0]
	if TForm do	SpLH.transform = inverse(WorkTM)
	SpLH.wirecolor = SpCL.wirecolor
	addnewspline SpLH
	addknot SpLH 1 #corner #line [0,0,0]
	addknot SpLH 1 #corner #line [0,0,0]
	
	-- REMOVE VISABILITY OF APPROPRIATE SPLINE
	case Just of
	(
		1: SpLH.ishidden = true
		2: SpCL.ishidden = true
		3: SpRH.ishidden = true
	)
		
	format prompts[LPrompt = 2]
	
	-- LOOP FOR POINTS IN MLINE SEQUENCE ***********************************************************************************************
	while CarryOn do
	(
		R = pickpoint terminators:#() mouseMoveCallback:MlineHdlr
		
		if (ClassOf R != Array) do 
		(
			delete #(SpRH,SpLH,SpCL)
			EscapeEnable=true
			SetListenerSel #(-1,-1)
			return (format "*Cancel*\nOK\n")
		)
		
		-- SET RETURN CODE	
		case of
		(
			(R[1]==#RightClick and R[2]==undefined):		RtnCode = 4		-- Right Mouse Click
			(Classof R[1]==Point3 and R[2]==undefined):		RtnCode = 1		-- Regular point input
			(R[1]=="c" and R[2]==undefined):				RtnCode = 2 	-- Close mline
			(R[1]=="C" and R[2]==undefined):				RtnCode = 2		-- Close mline
			(R[1]=="u" and R[2]==undefined):				RtnCode = 3		-- Undo
			(R[1]=="U" and R[2]==undefined):				RtnCode = 3		-- Undo
			(R[1]==#Escape and R[2]==undefined):			RtnCode = 6		-- Escape key
			default:										RtnCode = 4		
		)
		
		case RtnCode of
		(
			1:	(
					-- TEST FOR COMMON VIEWPORT	
					if (VT = view_type()) != WorkVP then 
					(
						msgString = "Input must be contained within the starting viewport, in this case the " + (WorkVP as string) + " viewport."
						messagebox  msgString
					)
					else
					(
						PT = PointConditioner testpt:R[1] tolerance:TolVal
									
						-- GET NUMBER OF KNOTS
						nK = NumKnots SpCL 1
						setKnotPoint SpCL 1 nK PT
						
						if nK == 2 do
						(
							CL1 = getKnotpoint SpCL 1 (nK - 1)
							CL2 = PT
							
							if TForm do
							(
								CL1 = PointConditioner testpt:(CL1 * WorkTM) tolerance:TolVal
								CL2 = PointConditioner testpt:(CL2 * WorkTM) tolerance:TolVal
							)
	
							lhRes = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
							rhRes = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
							
							if TForm then
							(
								LH1 = (PointConditioner testpt:(lhRes[1] * inverse(WorkTM)) tolerance:TolVal)
								LH2 = (PointConditioner testpt:(lhRes[2] * inverse(WorkTM)) tolerance:TolVal)
								RH1 = (PointConditioner testpt:(rhRes[1] * inverse(WorkTM)) tolerance:TolVal)
								RH2 = (PointConditioner testpt:(rhRes[2] * inverse(WorkTM)) tolerance:TolVal)
							)
							else
							(
								LH1 = lhRes[1]
								LH2 = lhRes[2]
								RH1 = rhRes[1]
								RH2 = rhRes[2]
							)
							
							setknotpoint SpLH 1 (nK-1) 	LH1
							setKnotPoint SpLH 1 (nK)	LH2
							setknotpoint SpRH 1 (nK-1) 	RH1
							setKnotPoint SpRH 1 (nK)	RH2

							updateshape SpLH
							updateshape SpCL
							updateshape SpRH

							-- ADD NEW KNOTS IN SEQUENCE: TEMPORARILY PLACE AT SAME POINTS AS PREVIOUS KNOTS
							addknot SpLH 1 #corner #line LH2
							addknot SpRH 1 #corner #line RH2
							addknot SpCL 1 #corner #line PT
							
							redrawviews()
						)
	
						if nK > 2 then
						(
							CL1 = getKnotpoint SpCL 1 (nK - 2)
							CL2 = getKnotpoint SpCL 1 (nK - 1)
							CL3 = PT
							
							if TForm then 
							(
								CL1 = PointConditioner testpt:(CL1 * WorkTM) tolerance:TolVal
								CL2 = PointConditioner testpt:(CL2 * WorkTM) tolerance:TolVal
								CL3 = PointConditioner testpt:(CL3 * WorkTM) tolerance:TolVal
							)
						
							rhRes1 = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
							rhRes2 = ParLine p1:CL2 p2:CL3 offset:RhOffset side:#right
							lhRes1 = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
							lhRes2 = ParLine p1:CL2 p2:CL3 offset:LhOffset side:#left
							aR = rhRes1[1]	;	aL = lhRes1[1]
							bR = rhRes1[2]	;	bL = lhRes1[2]
							cR = rhRes2[1]	;	cL = lhRes2[1]
							dR = rhRes2[2]	;	dL = lhRes2[2]

							RH2 = MLineIntersect [aR.x,aR.y] [bR.x,bR.y] [cR.x,cR.y] [dR.x,dR.y]
							RH3 = dR
							LH2 = MLineIntersect [aL.x,aL.y] [bL.x,bL.y] [cL.x,cL.y] [dL.x,dL.y]
							LH3 = dL
							
							if TForm do 
							(
								RH2 = PointConditioner testpt:(RH2 * inverse(WorkTM)) tolerance:TolVal
								RH3 = PointConditioner testpt:(dR * inverse(WorkTM)) tolerance:TolVal
								LH2 = PointConditioner testpt:(LH2 * inverse(WorkTM)) tolerance:TolVal
								LH3 = PointConditioner testpt:(dL * inverse(WorkTM)) tolerance:TolVal
							)
							
							setKnotpoint SpRH 1 (nK-1) RH2
							setKnotpoint SpRH 1 (nK) RH3
							setKnotpoint SpLH 1 (nK-1) LH2
							setKnotpoint SpLH 1 (nK) LH3
							
							updateshape SpLH
							updateshape SpRH
							updateshape SpCL
											
							-- ADD NEW KNOTS IN SEQUENCE: TEMPORARILY PLACE AT SAME POINTS AS PREVIOUS KNOTS
							addknot SpLH 1 #corner #line LH3
							addknot SpRH 1 #corner #line RH3
							addknot SpCL 1 #corner #line PT

							redrawviews()
						)
						
			
					)
					
					-- SEND NEXT PROMPT TO LISTENER
					case of
					(
						(nK == 2): ( format prompts[3] ; LPrompt = 3 )
						(nK >= 3): ( format prompts[4] ; LPrompt = 4 )
					)
				)
				
			2:	( -- CLOSE MLINE, NOTE THAT TWO SETS OF INTERSECTIONS HAVE TO BE DETERMINED FOR CLOSE OPERATION
					nK = numKnots SpCL 1
					if nK > 3 then
					(
						-- GET CENTERLINE POINTS
						CL1 = getKnotPoint SpCL 1 (nK-2)	
						CL2 = getKnotPoint SpCL 1 (nK-1)	
						CL3 = getKnotPoint SpCL 1 1			-- Original Point
						CL4 = getKnotPoint SpCL 1 2			-- 2nd Point
						
						-- SET FINAL KNOT ON CENTERLINE
						setKnotPoint SpCL 1 nK CL3
						
						-- TRANSFORM POINTS IF REQUIRED
						if TForm do
						(
							CL1 = PointConditioner testpt:(CL1*WorkTM) tolerance:TolVal
							CL2 = PointConditioner testpt:(CL2*WorkTM) tolerance:TolVal
							CL3 = PointConditioner testpt:(CL3*WorkTM) tolerance:TolVal
							CL4 = PointConditioner testpt:(CL4*WorkTM) tolerance:TolVal
						)
	
						-- GET POINTS DEFINING PARALLEL LINES TO THE LEFT AND RIGHT
						rhRes1 = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
						rhRes2 = ParLine p1:CL2 p2:CL3 offset:RhOffset side:#right
						rhRes3 = ParLine p1:CL3 p2:CL4 offset:RhOffset side:#right
				
						lhRes1 = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
						lhRes2 = ParLine p1:CL2 p2:CL3 offset:LhOffset side:#left
						lhRes3 = ParLine p1:CL3 p2:CL4 offset:LhOffset side:#left

						-- FIND INTERSECTION POINTS
						aR = rhRes1[1]	-- 1st line startpt 
						bR = rhRes1[2]	-- 1st line endpoint
						cR = rhRes2[1]	-- 2nd line startpt
						dR = rhRes2[2]	-- 2nd line endpoint
						fR = rhRes3[1]	-- 3rd line startpt
						gR = rhRes3[2]	-- 3rd line endpoint
					
						aL = lhRes1[1]	-- 1st line startpt 
						bL = lhRes1[2]	-- 1st line endpoint
						cL = lhRes2[1]	-- 2nd line startpt
						dL = lhRes2[2]	-- 2nd line endpoint
						fL = lhRes3[1]	-- 3rd line startpt
						gL = lhRes3[2]	-- 3rd line endpoint
						
						RH1 = MLineIntersect [aR.x,aR.y] [bR.x,bR.y] [cR.x,cR.y] [dR.x,dR.y]
						RH2 = MLineIntersect [cR.x,cR.y] [dR.x,dR.y] [fR.x,fR.y] [gR.x,gR.y]

						LH1 = MLineIntersect [aL.x,aL.y] [bL.x,bL.y] [cL.x,cL.y] [dL.x,dL.y]
						LH2 = MLineIntersect [cL.x,cL.y] [dL.x,dL.y] [fL.x,fL.y] [gL.x,gL.y]

						-- TRANSFORM POINTS BACK IF REQUIRED
						if TForm do
						(
							RH1 = PointConditioner testpt:(RH1*inverse(WorkTM)) tolerance:TolVal
							RH2 = PointConditioner testpt:(RH2*inverse(WorkTM)) tolerance:TolVal
							LH1 = PointConditioner testpt:(LH1*inverse(WorkTM)) tolerance:TolVal
							LH2 = PointConditioner testpt:(LH2*inverse(WorkTM)) tolerance:TolVal
 						)
						
						-- SET RIGHT AND LEFT KNOT POINTS
						setKnotPoint SpRH 1 1 RH2
						setKnotPoint SpRH 1 (nK-1) RH1
						setKnotPoint SpLH 1 1 LH2
						setKnotPoint SpLH 1 (nK-1) LH1
						
						-- DELETE LAST KNOTPOINTS, THESE WERE USED FOR CALCULATION ONLY
						deleteKnot SpRH 1 nK
						deleteKnot SpLH 1 nK
						deleteKnot SpCL 1 nK
						
						-- CLOSE EACH OF THE MLINE SEGMENTS
						Close SpCL 1
						Close SpRH 1
						Close SpLH 1
												
						-- UPDATE SHAPE CACHE AND REDRAW
						updateshape SpCL
						updateshape SpRH
						updateshape SpLH
						redrawviews()
					)
					
					-- DELETE APPROPRIATE SPLINE
					case Just of
					(
						1: delete SpLH
						2: delete SpCL
						3: delete SpRH
					)
					
					-- PRINT FINAL STRING TO LISTENER
					format prompts[5]
					CarryOn = false
				)
			
			3:	( -- UNDO
					nK = numKnots SpCL 1
					
					if nK == 3 do
					(
						deleteKnot SpCL 1 (nK-1)
						deleteKnot SpRH 1 (nK-1)
						deleteKnot SpLH 1 (nK-1)
					
						nK = numKnots SpCL 1
						
						CL1 = getKnotpoint SpCL 1 (nK - 1)
						CL2 = getKnotPoint SpCL 1 nK
					
						if TForm then 
						(
							CL1 = PointConditioner testpt:(CL1 * WorkTM) tolerance:TolVal
							CL2 = PointConditioner testpt:(CL2 * WorkTM) tolerance:TolVal
						)
						
						rhRes = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
						lhRes = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
						
						RH1 = rhRes[1]
						RH2 = rhRes[2]
						LH1 = lhRes[1]
						LH2 = lhRes[2]
						
						if TForm do 
						(
							RH1 = PointConditioner testpt:(RH1 * inverse(WorkTM)) tolerance:TolVal
							RH2 = PointConditioner testpt:(RH2 * inverse(WorkTM)) tolerance:TolVal
							LH1 = PointConditioner testpt:(LH1 * inverse(WorkTM)) tolerance:TolVal
							LH2 = PointConditioner testpt:(LH2 * inverse(WorkTM)) tolerance:TolVal
						)

						setKnotpoint SpRH 1 (nK-1) 	RH1
						setKnotpoint SpRH 1 (nK) 	RH2
						setKnotpoint SpLH 1 (nK-1) 	LH1
						setKnotpoint SpLH 1 (nK) 	LH2
	
						updateshape SpLH
						updateshape SpRH
						updateshape SpCL
											
						redrawviews()
					)

					if nK > 3 do
					(
						deleteKnot SpCL 1 (nK-1)
						deleteKnot SpRH 1 (nK-1)
						deleteKnot SpLH 1 (nK-1)
						
						nK = numKnots SpCL 1
					
						CL1 = getKnotpoint SpCL 1 (nK - 2)
						CL2 = getKnotpoint SpCL 1 (nK - 1)
						CL3 = getKnotPoint SpCL 1 nK
													
						if TForm then 
						(
							CL1 = PointConditioner testpt:(CL1 * WorkTM) tolerance:TolVal
							CL2 = PointConditioner testpt:(CL2 * WorkTM) tolerance:TolVal
							CL3 = PointConditioner testpt:(CL3 * WorkTM) tolerance:TolVal
						)
					
						rhRes1 = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
						rhRes2 = ParLine p1:CL2 p2:CL3 offset:RhOffset side:#right
						lhRes1 = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
						lhRes2 = ParLine p1:CL2 p2:CL3 offset:LhOffset side:#left
						
						aR = rhRes1[1]	;	aL = lhRes1[1]
						bR = rhRes1[2]	;	bL = lhRes1[2]
						cR = rhRes2[1]	;	cL = lhRes2[1]
						dR = rhRes2[2]	;	dL = lhRes2[2]
		
						RH2 = MLineIntersect [aR.x,aR.y] [bR.x,bR.y] [cR.x,cR.y] [dR.x,dR.y]
						RH3 = dR
						LH2 = MLineIntersect [aL.x,aL.y] [bL.x,bL.y] [cL.x,cL.y] [dL.x,dL.y]
						LH3 = dL
						
						if TForm do 
						(
							RH2 = PointConditioner testpt:(RH2 * inverse(WorkTM)) tolerance:TolVal
							RH3 = PointConditioner testpt:(dR * inverse(WorkTM)) tolerance:TolVal
							LH2 = PointConditioner testpt:(LH2 * inverse(WorkTM)) tolerance:TolVal
							LH3 = PointConditioner testpt:(dL * inverse(WorkTM)) tolerance:TolVal
						)
						
						setKnotpoint SpRH 1 (nK-1) RH2
						setKnotpoint SpRH 1 (nK) RH3
						setKnotpoint SpLH 1 (nK-1) LH2
						setKnotpoint SpLH 1 (nK) LH3
						
						updateshape SpLH
						updateshape SpRH
						updateshape SpCL
											
						redrawviews()
					)
						
					-- SEND NEXT PROMPT TO LISTENER
					format "\n"
					case of
					(
						(nK == 3): ( format prompts[2] ; LPrompt = 2 )
						(nK == 4): ( format prompts[3] ; LPrompt = 3 )
						(nK >= 5): ( format prompts[4] ; LPrompt = 4 )
					)
				)
				
			4:	( -- RIGHT MOUSE CLICK
					
					-- Get last key entered
					CaretPos = getListenerSel()
					setListenerSel #((CaretPos[1]-1),CaretPos[2])
					SelText = getListenerSelText()
					case of
					(
						(SelText=="C" or SelText=="c"): 
						(
							nK = numKnots SpCL 1
							if nK > 3 then
							(
								-- GET CENTERLINE POINTS
								CL1 = getKnotPoint SpCL 1 (nK-2)	
								CL2 = getKnotPoint SpCL 1 (nK-1)	
								CL3 = getKnotPoint SpCL 1 1			-- Original Point
								CL4 = getKnotPoint SpCL 1 2			-- 2nd Point
								
								-- SET FINAL KNOT ON CENTERLINE
								setKnotPoint SpCL 1 nK CL3
								
								-- TRANSFORM POINTS IF REQUIRED
								if TForm do
								(
									CL1 = PointConditioner testpt:(CL1*WorkTM) tolerance:TolVal
									CL2 = PointConditioner testpt:(CL2*WorkTM) tolerance:TolVal
									CL3 = PointConditioner testpt:(CL3*WorkTM) tolerance:TolVal
									CL4 = PointConditioner testpt:(CL4*WorkTM) tolerance:TolVal
								)
			
								-- GET POINTS DEFINING PARALLEL LINES TO THE LEFT AND RIGHT
								rhRes1 = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
								rhRes2 = ParLine p1:CL2 p2:CL3 offset:RhOffset side:#right
								rhRes3 = ParLine p1:CL3 p2:CL4 offset:RhOffset side:#right
						
								lhRes1 = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
								lhRes2 = ParLine p1:CL2 p2:CL3 offset:LhOffset side:#left
								lhRes3 = ParLine p1:CL3 p2:CL4 offset:LhOffset side:#left
		
								-- FIND INTERSECTION POINTS
								aR = rhRes1[1]	-- 1st line startpt 
								bR = rhRes1[2]	-- 1st line endpoint
								cR = rhRes2[1]	-- 2nd line startpt
								dR = rhRes2[2]	-- 2nd line endpoint
								fR = rhRes3[1]	-- 3rd line startpt
								gR = rhRes3[2]	-- 3rd line endpoint
							
								aL = lhRes1[1]	-- 1st line startpt 
								bL = lhRes1[2]	-- 1st line endpoint
								cL = lhRes2[1]	-- 2nd line startpt
								dL = lhRes2[2]	-- 2nd line endpoint
								fL = lhRes3[1]	-- 3rd line startpt
								gL = lhRes3[2]	-- 3rd line endpoint
								
								RH1 = MLineIntersect [aR.x,aR.y] [bR.x,bR.y] [cR.x,cR.y] [dR.x,dR.y]
								RH2 = MLineIntersect [cR.x,cR.y] [dR.x,dR.y] [fR.x,fR.y] [gR.x,gR.y]
		
								LH1 = MLineIntersect [aL.x,aL.y] [bL.x,bL.y] [cL.x,cL.y] [dL.x,dL.y]
								LH2 = MLineIntersect [cL.x,cL.y] [dL.x,dL.y] [fL.x,fL.y] [gL.x,gL.y]
		
								-- TRANSFORM POINTS BACK IF REQUIRED
								if TForm do
								(
									RH1 = PointConditioner testpt:(RH1*inverse(WorkTM)) tolerance:TolVal
									RH2 = PointConditioner testpt:(RH2*inverse(WorkTM)) tolerance:TolVal
									LH1 = PointConditioner testpt:(LH1*inverse(WorkTM)) tolerance:TolVal
									LH2 = PointConditioner testpt:(LH2*inverse(WorkTM)) tolerance:TolVal
		 						)
								
								-- SET RIGHT AND LEFT KNOT POINTS
								setKnotPoint SpRH 1 1 RH2
								setKnotPoint SpRH 1 (nK-1) RH1
								setKnotPoint SpLH 1 1 LH2
								setKnotPoint SpLH 1 (nK-1) LH1
								
								-- DELETE LAST KNOTPOINTS, THESE WERE USED FOR CALCULATION ONLY
								deleteKnot SpRH 1 nK
								deleteKnot SpLH 1 nK
								deleteKnot SpCL 1 nK
								
								-- CLOSE EACH OF THE MLINE SEGMENTS
								Close SpCL 1
								Close SpRH 1
								Close SpLH 1
														
								-- UPDATE SHAPE CACHE AND REDRAW
								updateshape SpCL
								updateshape SpRH
								updateshape SpLH
								redrawviews()
							)
							
							-- DELETE APPROPRIATE SPLINE
							case Just of
							(
								1: delete SpLH
								2: delete SpCL
								3: delete SpRH
							)
							
							-- PRINT FINAL STRING TO LISTENER
							format prompts[5]
							CarryOn = false
						)
						(SelText=="U" or SelText=="u"):
						(
							nK = numKnots SpCL 1
							
							if nK == 3 do
							(
								deleteKnot SpCL 1 (nK-1)
								deleteKnot SpRH 1 (nK-1)
								deleteKnot SpLH 1 (nK-1)
							
								nK = numKnots SpCL 1
								
								CL1 = getKnotpoint SpCL 1 (nK - 1)
								CL2 = getKnotPoint SpCL 1 nK
							
								if TForm then 
								(
									CL1 = PointConditioner testpt:(CL1 * WorkTM) tolerance:TolVal
									CL2 = PointConditioner testpt:(CL2 * WorkTM) tolerance:TolVal
								)
								
								rhRes = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
								lhRes = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
								
								RH1 = rhRes[1]
								RH2 = rhRes[2]
								LH1 = lhRes[1]
								LH2 = lhRes[2]
								
								if TForm do 
								(
									RH1 = PointConditioner testpt:(RH1 * inverse(WorkTM)) tolerance:TolVal
									RH2 = PointConditioner testpt:(RH2 * inverse(WorkTM)) tolerance:TolVal
									LH1 = PointConditioner testpt:(LH1 * inverse(WorkTM)) tolerance:TolVal
									LH2 = PointConditioner testpt:(LH2 * inverse(WorkTM)) tolerance:TolVal
								)
		
								setKnotpoint SpRH 1 (nK-1) 	RH1
								setKnotpoint SpRH 1 (nK) 	RH2
								setKnotpoint SpLH 1 (nK-1) 	LH1
								setKnotpoint SpLH 1 (nK) 	LH2
			
								updateshape SpLH
								updateshape SpRH
								updateshape SpCL
													
								redrawviews()
							)
		
							if nK > 3 do
							(
								deleteKnot SpCL 1 (nK-1)
								deleteKnot SpRH 1 (nK-1)
								deleteKnot SpLH 1 (nK-1)
								
								nK = numKnots SpCL 1
							
								CL1 = getKnotpoint SpCL 1 (nK - 2)
								CL2 = getKnotpoint SpCL 1 (nK - 1)
								CL3 = getKnotPoint SpCL 1 nK
															
								if TForm then 
								(
									CL1 = PointConditioner testpt:(CL1 * WorkTM) tolerance:TolVal
									CL2 = PointConditioner testpt:(CL2 * WorkTM) tolerance:TolVal
									CL3 = PointConditioner testpt:(CL3 * WorkTM) tolerance:TolVal
								)
							
								rhRes1 = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
								rhRes2 = ParLine p1:CL2 p2:CL3 offset:RhOffset side:#right
								lhRes1 = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
								lhRes2 = ParLine p1:CL2 p2:CL3 offset:LhOffset side:#left
								
								aR = rhRes1[1]	;	aL = lhRes1[1]
								bR = rhRes1[2]	;	bL = lhRes1[2]
								cR = rhRes2[1]	;	cL = lhRes2[1]
								dR = rhRes2[2]	;	dL = lhRes2[2]
				
								RH2 = MLineIntersect [aR.x,aR.y] [bR.x,bR.y] [cR.x,cR.y] [dR.x,dR.y]
								RH3 = dR
								LH2 = MLineIntersect [aL.x,aL.y] [bL.x,bL.y] [cL.x,cL.y] [dL.x,dL.y]
								LH3 = dL
								
								if TForm do 
								(
									RH2 = PointConditioner testpt:(RH2 * inverse(WorkTM)) tolerance:TolVal
									RH3 = PointConditioner testpt:(dR * inverse(WorkTM)) tolerance:TolVal
									LH2 = PointConditioner testpt:(LH2 * inverse(WorkTM)) tolerance:TolVal
									LH3 = PointConditioner testpt:(dL * inverse(WorkTM)) tolerance:TolVal
								)
								
								setKnotpoint SpRH 1 (nK-1) RH2
								setKnotpoint SpRH 1 (nK) RH3
								setKnotpoint SpLH 1 (nK-1) LH2
								setKnotpoint SpLH 1 (nK) LH3
								
								updateshape SpLH
								updateshape SpRH
								updateshape SpCL
													
								redrawviews()
							)
								
							-- SEND NEXT PROMPT TO LISTENER
							format "\n"
							case of
							(
								(nK == 3): ( format prompts[2] ; LPrompt = 2 )
								(nK == 4): ( format prompts[3] ; LPrompt = 3 )
								(nK >= 5): ( format prompts[4] ; LPrompt = 4 )
							)
						)
						default:
						(
							nK = numKnots SpCL 1
							if nK == 2 then
							(
								delete #(SpRH,SpLH,SpCL)
							)
							else
							(			
								deleteKnot SpCL 1 nK
								deleteKnot SpRH 1 nK
								deleteKnot SpLH 1 nK
						
								CL1 = getKnotpoint SpCL 1 (nK - 2)
								CL2 = getKnotpoint SpCL 1 (nK - 1)
													
								if TForm then 
								(
									CL1 = PointConditioner testpt:(CL1 * WorkTM) tolerance:TolVal
									CL2 = PointConditioner testpt:(CL2 * WorkTM) tolerance:TolVal
								)
							
								rhRes1 = ParLine p1:CL1 p2:CL2 offset:RhOffset side:#right
								lhRes1 = ParLine p1:CL1 p2:CL2 offset:LhOffset side:#left
												
								RH2 = rhRes1[2]
								LH2 = lhRes1[2]
								
								if TForm do 
								(
									RH2 = PointConditioner testpt:(RH2 * inverse(WorkTM)) tolerance:TolVal
									LH2 = PointConditioner testpt:(LH2 * inverse(WorkTM)) tolerance:TolVal
								)
								
								setKnotpoint SpRH 1 (nK-1) RH2
								setKnotpoint SpLH 1 (nK-1) LH2
													
								updateshape SpLH
								updateshape SpRH
								updateshape SpCL
							)
								
							redrawviews()
							CarryOn = false
							format prompts[5]
							LPrompt = 5
						)
					)
				)
				
			6:	( 	-- ESCAPE KEY PRESSED
					delete #(SpRH,SpLH,SpCL)
					EscapeEnable=true
					setListenerSel #(-1,-1)
					return(format "*Cancel*\nOK\n")
				) 

			default:
			(
				EscapeEnable=true
				setListenerSel #(-1,-1)
				return(format "*Cancel*\nOK\n")
			)		
		) 		
	) 
	
	set coordsys CS
	EscapeEnable = true
)

-- MLINE FUNCTION ***************************************************************************
fn _mline = 
(
	if (ClassOf AcadFloat1 == RollOutFloater) do closeRolloutFloater AcadFloat1

	AcadFloat1 = newrolloutfloater "Mline" 420 210		
	addrollout roMline AcadFloat1
)

