-- AUTOCAD PLINE COMMAND EMULATION

-- ************************************************************************************
-- This function returns the tangent "in vector" direction
-- associated with the last knot in a splineshape
fn GetLastTangent id = 
(
	-- DECLARE LOCALS
	local nK
	local p1,p2,pt
	local CS
	
	CS = set coordsys world
	
	-- TEST FOR VALIDITY OF ID ARGUMENT
	if classof id != splineshape do return false
	
	-- GET THE TANGENT DIRECTION
	nK = numKnots id 1
	p1 = getInVec id 1 nK
	p2 = getKnotPoint id 1 nK
	pt = normalize (p2 - p1)
	
	set coordsys CS
	return pt
)

-- ****************************************************************************************
fn getLastPoint id =
(
	-- DECLARE LOCALS
	local nK
	local p1,p2
	local CS
	
	CS = set coordsys world
	
	-- TEST FOR VALIDITY OF ID ARGUMENT
	if classof id != splineshape do return false
	
	-- GET THE TANGENT DIRECTION
	nK = numKnots id 1
	pt = getKnotPoint id 1 nK
	
	set coordsys CS
	return pt
)

-- **********************************************************************************
fn getMidPoint pt1 pt2 =
(
	local pt,MPt

	pt = pt1 + pt2	
	MPt = [pt.x/2.0, pt.y/2.0, 0]
)

-- **********************************************************************************
fn LineIntersect p1 p2 q1 q2 =
(
	-- VARIABLE DECLARATIONS
	local A,B,C
	local p,r,Denom
	
	-- SETUP FOR INTERSECTION POINT CALCULATION
	A = p2-p1
	B = q1-q2
	C = q1-p1
	
	-- DENOMINATOR NEAR 0 IMPLIES LINES APPROX PARALLEL
	if abs(Denom = A.y*B.x-B.y*A.x) < 0.0001 then
	(	
		[(p2.x+q1.x)/2.0,(p2.y+q1.y)/2.0,0]
	)	
	else
	(	
		-- PARAMETRIC CONSTANTS FROM CRAMER'S RULE
		-- r is associated with line p1p2
		r = (C.y*B.x-B.y*C.x) / Denom
	
		-- s is associated with line q1q2, not required here but leave in for
		-- possible future additions.
		-- s = (A.y*C.x-C.y*A.x) / Denom	
	
		-- DETERMINE INTERSECTION POINT
		p = p1 + r * A
		[p.x,p.y,0.0]
	)
)

-- GET ARC DATA FUNCTION *****************************************************
-- Parameters: 
-- direction is the normalized starting direction of the arc (point3)	
-- start is the start point of the arc (point3)
-- end is the endpoint of the arc (point3)		
-- Returns array containing arc parameters:
-- Arc center point, Arc radius, Arc starting angle, arc ending angle		
fn getArcData direction: start: end: =
(
	local Band,TestVec,VSide
	local BandMidPt,DirectionPerp,BandPerp
	local ArcCenPt,ArcRad
	local StartVector,EndVector
	local Angle1,Angle2,FromAngle,ToAngle
	
	-- NORMALIZE RUBBERBAND VECTOR
	Band = normalize (end - start)
		
	-- TEST WHETHER RUBBERBAND GOES RIGHT OR LEFT OF LAST TANGENT DIRECTION
	TestVec = cross [direction.x,direction.y,0] [band.x,band.y,0]
	if TestVec.z > 0.0 
	then	VSide = #LEFT
	else	VSide = #RIGHT
				
	-- FIND MIDPOINT OF RUBBERBAND
	BandMidPt = start + 0.5 * (end - start)
						
	-- PERPENDICULARS TO LAST SEGMENT AND RUBBERBAND  
	DirectionPerp = cross direction z_axis
	BandPerp = cross Band z_axis
								
	-- CENTER POINT AND RADIUS OF ARC
	
	arg1 = start
	arg2 = start + DirectionPerp
	arg3 = BandMidPt
	arg4 = BandMidPt + BandPerp
	
	ArcCenPt = LineIntersect arg1 arg2 arg3 arg4
	ArcRad = length (ArcCenPt - start)
		
	-- DETERMINE FROM AND TO ANGLES OF ARC
	StartVector = (start - ArcCenPt)
	EndVector = (end - ArcCenPt)
			
	Angle1 = atan2 StartVector.y StartVector.x
	if Angle1 < 0 do Angle1 = 360 + Angle1
			
	Angle2 = atan2 EndVector.y EndVector.x
	if Angle2 < 0 do Angle2 = 360 + Angle2

	if VSide == #Right then
	(
		FromAngle = Angle2
		ToAngle = Angle1
	)
	else
	(
		FromAngle = Angle1
		ToAngle = Angle2
	)
	
	-- RETURN ARC DATA
	#(ArcCenPt,ArcRad,FromAngle,ToAngle,VSide)
)


-- *****************************************************************************************
fn MakeArc id: endpoint: origin: trans: matrix: memory: =
(
	-- PARAMETER DESCRIPTIONS:
	-- id:			node identifier
	-- endpoint: 	endpoint of the arc to be created in world coordinates(point3)
	-- origin: 		original point on the pline in world coords (point3)
	-- trans: 		flag indicating whether points must be transformed from WCS before use (bool)
	-- matrix: 		World CS to Working CS transformation matrix, (matrix3)
	
	local ArcStartDir,ArcStart,ArcEnd,ArcData
	local TempArc,VSide
	local K1,K1in,K1out ,K2,K2in,K2out 
	local K3,K3in,K3out ,K4,K4in,K4out
	local CS
	local TolVal = 0.0005
	local ZLocal
	local DirMatrix
	local nK, TempVec
	
	ZLocal = matrix.row3
	
	CS = set coordsys world
	
	nK = numknots id 1
	
	-- GET TANGENT DIRECTION, LAST POINT AND ENDPOINT FOR PASSING TO ARC DATA FUNCTION
	if (nK == 0) then
	(	
		ArcStartDir = PointConditioner testpt:(cross ZLocal (normalize (endpoint - origin))) tolerance:TolVal
		ArcStart = PointConditioner testpt:origin tolerance:TolVal
	)
	else
	(
		ArcStartDir = PointConditioner testpt:(GetLastTangent id) tolerance:TolVal
		ArcStart = PointConditioner testpt:(GetLastPoint id) tolerance:TolVal
	)
	
	ArcEnd = endpoint
		
	-- TRANSFORM TO LOCAL C.S. IF REQUIRED
	if trans do
	(
		-- Note: ArcStartDir is a vector so its orientation 
		-- requires transformation, however it does not require translation.
		DirMatrix = matrix3 matrix.row1 matrix.row2 matrix.row3 [0,0,0]
		ArcStartDir = PointConditioner testpt:(ArcStartDir * DirMatrix) tolerance: TolVal
		
		ArcStart = PointConditioner testpt:(ArcStart * matrix) tolerance:TolVal
		ArcEnd = PointConditioner testpt:(ArcEnd * matrix) tolerance:TolVal
	)
	
	
	-- GET ARC DATA, DIRECTION, ALL PARAMETERS MUST BE EXPRESSED IN WORKING C.S. BY THIS POINT
	ArcData = getArcData direction:ArcStartDir start:ArcStart end:ArcEnd

	-- BUILD TEMPORARY ARC TO DERIVE DEFINITION DATA 
	TempArc = arc pos:ArcData[1] radius:ArcData[2] from:ArcData[3] to:ArcData[4]
	converttosplineshape TempArc
	
	-- DERIVE DATA
	K1 = getKnotPoint TempArc 1 1
	K1in = getInVec TempArc 1 1
	K1out = getOutVec TempArc 1 1
	K2 = getKnotPoint TempArc 1 2
	K2in = getInVec TempArc 1 2
	K2out = getOutVec TempArc 1 2
	K3 = getKnotPoint TempArc 1 3
	K3in = getInVec TempArc 1 3
	K3out = getOutVec TempArc 1 3
	K4 = getKnotPoint TempArc 1 4
	K4in = getInVec TempArc 1 4
	K4out = getOutVec TempArc 1 4
	
	-- DELETE TEMPORARY ARC SPLINESHAPE
	delete TempArc
	
	VSide = ArcData[5]
	
	if trans do (set coordsys (inverse matrix))
	
	if VSide == #Left then
	(
		if (nK > 0) then 
		(
			TempVec = getInVec id 1 nK 
			deleteKnot id 1 nK
			if (memory[(memory.count)]==#ARC) 
			then addknot id 1 #bezier #curve K1 tempVec K1out
			else addknot id 1 #bezier #curve K1 K1in K1out
		)
		else addknot id 1 #bezier #curve K1 K1in K1out
	
		addknot id 1 #bezier #curve K2 K2in K2out
		addknot id 1 #bezier #curve K3 K3in K3out
		addknot id 1 #bezier #curve K4 K4in K4out
	)
	else
	(
		if (nK > 0) then
		(
			TempVec = getInVec id 1 nK
			deleteKnot id 1 nK
			if (memory[(memory.count)]==#ARC)
			then addknot id 1 #bezier #curve K4 tempVec K4in
			else addknot id 1 #bezier #curve K4 K4out K4in
		)
		else addknot id 1 #bezier #curve K4 K4out K4in
		
		addknot id 1 #bezier #curve K3 K3out K3in
		addknot id 1 #bezier #curve K2 K2out K2in
		addknot id 1 #bezier #curve K1 K1out K1in
	)
	updateshape id
	redrawviews()
	set coordsys CS
)

-- **********************************************************************************
fn MakeLine id: endpoint: origin: trans: matrix: = 
(
	local CS
	local TolVal = 0.0005

	CS = set coordsys world
	
	origin = PointConditioner testpt:origin  tolerance:TolVal
	endpoint = PointConditioner testpt:endpoint tolerance:TolVal
	
	
	if (numKnots id 1) == 0 then
	(
		addknot id 1 #corner #line origin
		addknot id 1 #corner #line endpoint
		updateshape id
		redrawviews()
	)
	else
	(
		addknot id 1 #corner #line endpoint
		updateshape id
		redrawviews()
	)
	set coordsys CS
)

-- *********************************************************************************************
fn UndoLine id: trans: matrix: =
(
	local nK,Pt
	local WireColor
	
	nK = numKnots id 1
	
	-- ONLY ONE LINE LEFT; DELETE AND CREATE NEW PLINE
	if nK == 2 then	
	(
		WireColor=id.wirecolor
		delete id
		id = splineshape pos:[0,0,0]
			
		addnewspline id
		id.wirecolor = WireColor
		if trans do ( id.transform = inverse(WorkTM) )
 		Pt = undefined
	)
	else
	(
		deleteKnot id 1 nK
		updateshape id
		Pt = getKnotPoint id 1 (nK-1)
	)
	#(id,Pt)
)

-- *************************************************************************************
fn UndoArc id: trans: matrix: memory: =
(
	local nK,Pt,TempPt
	local WireColor
	
	nK = numKnots id 1
	
	-- ONLY ONE ARC LEFT; DELETE AND CREATE NEW PLINE
	if nK == 4 then
	(	
		WireColor=id.wirecolor
		delete id
		id = splineshape pos:[0,0,0]
		id.wirecolor=WireColor
		addnewspline id
		if trans do ( id.trans = inverse(WorkTM) )
 		Pt = undefined
	)
	else
	(
		deleteKnot id 1 nK
		deleteKnot id 1 (nK-1)
		deleteKnot id 1 (nK-2)
		
		-- DETERMINE WHETHER PREVIOUS SEGMENT IS AN ARC OR LINE
		if (memory[(memory.count-1)] == #LINE ) then
		(-- LINE
			TempPt = getKnotPoint id 1 (nK-3)
			deleteKnot id 1 (nK-3)
			addKnot id 1 #corner #line TempPt
		)
		else
		(-- ARC
			setOutVec id 1 (nK-3) (getKnotPoint id 1 (nK-3))
		)
		Pt = (getKnotPoint id 1 (nK-3))
		updateshape id
	)
	#(id,Pt)
)


-- MAIN PLINE FUNCTION **************************************************************	
fn _pline =
(

	-- VARIABLE DECLARATIONS
	local PItem
	local MArcID = 0
	local MCall = false
	local pt,P1,P2
	local DrawMode = #LINE
	local CarryOn = true
	local R
	local VSide
	local Prompts = #()
	local TolVal = 0.0005
	local WorkTM,TForm,IWorkTM,ZLocal
	local ViewType,CS
	local ArcData
	local StartWithArc = false
	local a,b
	local SelText,CaretPos
	local UndoRtnVal
	local TypeRecord = #()
	local WireColor
		
	-- MOUSE MOVE CALLBACK HANDLER 
	fn MHandler K4 =
	(
		-- In this handler all coordinate values are transformed to the working coordinate system
				
		local ArcStartDir,ArcStart,ArcEnd
		local ArcData,TM
		local CS,DirMatrix
		
		-- ONLY TAKE ACTION IF DRAWMODE = ARC, (RUBBERBAND HANDLES OTHERWISE)
		if _pline.DrawMode == #ARC do
		(
			-- RETURN IF THERE ARE NO SPLINES ASSIGNED TO PLINE YET
			--if PItem.numsplines == 0 do return true
						
			-- RETURN IF THE MOUSE HAS NOT MOVED MUCH
			if (length (k4-_pline.Pt)) <= .001 do return true
			
			CS = set coordsys world
						
			-- GET TANGENT DIRECTION, LAST POINT AND ENDPOINT FOR PASSING TO ARC DATA FUNCTION
			if _pline.StartWithArc then
			(
				ArcStartDir = (cross makeArc.ZLocal (normalize (K4 - _pline.P1)))
				ArcStart = _pline.P1
			)
			else
			(
				ArcStartDir = GetLastTangent _pline.PItem
				ArcStart = GetLastPoint _pline.PItem
			)
			
			ArcEnd = K4
						
			-- TRANSFORM TO WORKING C.S. IF REQUIRED 
			if _pline.TForm do
			(
				DirMatrix = Matrix3 _pline.WorkTM.row1 _pline.WorkTM.row2 _pline.WorkTM.row3 [0,0,0]
				ArcStartDir = ArcStartDir * DirMatrix
				ArcStart = ArcStart * _pline.WorkTM
				ArcEnd = ArcEnd * _pline.WorkTM
			)
			
			-- GET ARC DATA
			ArcData = getArcData direction:ArcStartDir start:ArcStart end:ArcEnd
			
			if not _pline.MCall then
			(
				_pline.MArcID = Arc pos:ArcData[1] radius:ArcData[2] from:ArcData[3] to:ArcData[4]
				_pline.MArcID.wirecolor = _pline.WireColor
				_pline.MCall = true
					
				if _pline.TForm do
				(
					TM = _pline.IWorkTM
					TM = Matrix3 TM.row1 TM.row2 TM.row3 (ArcData[1] * _pline.IWorkTM)
					_pline.MArcID.transform = TM 
				)
			)
			else
			(
				CS = set coordsys _pline.IWorkTM
				_pline.MArcID.pos = ArcData[1]
				_pline.MArcID.radius = ArcData[2]
				_pline.MArcID.from = ArcData[3]
				_pline.MArcID.to = ArcData[4]
				set coordsys CS
			)
			redrawviews()
		)
	)
	
	-- CODE BODY ***************************************************************************************************
	-- *************************************************************************************************************
	
	-- SYSTEM SETTINGS
	EscapeEnable = false
	ClearListener()
	
	-- SET SYSTEM PROMPTS
	Prompts[1] = "From point: \n"
	Prompts[2] = "Arc/Close/Undo/<Endpoint of line:> \n"
	Prompts[3] = "Close/Line/Undo/<Endpoint of arc:> \n"
	Prompts[4] = "To point: \n"
	Prompts[5] = "Arc/<Endpoint of line:> \n"

		
	-- GET FIRST POINT
	format Prompts[LPrompt = 1]
	Pt = P1 = pickpoint() 
	
	if (Pt == #Escape or Pt==#RightClick or classof P1 != Point3) do
	(
		EscapeEnable = true 
		setListenerSel #(-1,-1)
		return (format "*Cancel*\nOK\n" )
	)
	
	-- DETERMINE WHETHER USER GRID EXISTS AND IS ACTIVE
	a = getCPTM()	
	b = matrix3 1
	if (a.row1==b.row1 and a.row2==b.row2 and a.row3==b.row3 and a.row4==b.row4) 
	then UserGrid = false
	else UserGrid = true
	
	-- DETERMINE WHAT TRANSFORMATION IS REQD (IF NOT REQD SET TO WCS; MATRIX 3 1)
	if UserGrid 
	then 
	( 
		WorkTM = inverse a
		TForm = true
	)
	else 
	(
		case view_type() of 
		(
			#perspective: 	( TForm = false ; WorkTM = matrix3 1 )
			#user:			( TForm = false ; WorkTM = matrix3 1 )
			#top:			( Tform = false ; WorkTM = matrix3 1 )
			default:		( TForm = true ; WorkTM = getViewTM())
		)
	)

	-- SAVE VIEW TYPE FOR CHECKING USER INPUT
	ViewType = view_type()
	
	IWorkTM = inverse WorkTM
	ZLocal = WorkTM.row3
	
	-- CREATE EMPTY SPLINESHAPE AND SET TRANSFORM TO COINCIDE WITH WORK PLANE
	PItem = splineshape pos:[0,0,0]	
	addnewspline PItem
	
	-- Set WireColor Variable
	WireColor = PItem.wirecolor
	
	if TForm do ( PItem.transform = inverse(WorkTM) )	
		
	format Prompts[LPrompt = 5]
		
	-- LOOP TO COLLECT PLINE VERTICES
	while CarryOn do
	(
		R = pickpoint rubberband: pt terminators:#() MouseMoveCallBack: MHandler
		
		if ClassOf R != Array do 
		(
			delete PItem
			if MCall do delete MArcID
			max views redraw
			EscapeEnable=true
			setListenerSel #(-1,-1)
			return format "*Cancel*\nOK\n"
		)	
	 
		-- SET RETURN CODE	
		case of
		(
			(Classof R[1] == Point3 and R[2]==undefined):	RtnCode = 1		-- Regular point input
			(R[1]=="a"  and R[2]==undefined):				RtnCode = 2 	-- Switch to Arc Mode
			(R[1]=="A" and R[2]==undefined):				RtnCode = 2 	-- Switch to Arc Mode
			(R[1]=="l" and R[2]==undefined):				RtnCode = 3		-- Switch to Line Mode
			(R[1]=="L" and R[2]==undefined):				RtnCode = 3 	-- Switch to Line Mode
			(R[1]=="c" and R[2]==undefined):				RtnCode = 4 	-- Close pline
			(R[1]=="C" and R[2]==undefined):				RtnCode = 4		-- Close pline
			(R[1]=="u" and R[2]==undefined):				RtnCode = 5		-- Undo
			(R[1]=="U" and R[2]==undefined):				RtnCode = 5		-- Undo
			(R[1]=="" and R[2]==undefined):					RtnCode = 6		-- Enter key press
			(R[1]==#Escape and R[2]==undefined):			RtnCode = 8		-- Escape key
			(R[1]==#RightClick and R[2]==undefined):		RtnCode = 9		-- Right Mouse Click
			default:										RtnCode = 6		-- Exit command
		)
		
		-- TAKE APPROPRIATE ACTION BASED ON RETURN CODE
		case RtnCode of
		(
			1: 	( 	-- CHECK FOR INPUT IN CORRECT VIEWPORT
					if (view_type() != ViewType) then 
					(
						msgString = "Restrict input to " + ViewType + " viewport."
						messagebox msgString
						delete PItem
						max views redraw
						setListenerSel #(-1,-1)
						return (format "*Cancel*\nOK\n")
					)
					else -- POINT INPUT OK, MAKE NEW LINE OR ARC DEPENDING ON STATE OF DRAWMODE VARIABLE
					(
						-- CONDITION RETURNED POINT
						P2 = PointConditioner testpt:R[1] tolerance:TolVal
						
						if StartWithArc do (StartWithArc = false)
						
						if DrawMode == #LINE then 
						(
							MakeLine id:PItem endpoint:P2 origin:P1 transform:TForm matrix:WorkTM
							append TypeRecord #LINE
							format Prompts[LPrompt = 2]
						)
						else -- DrawMode = ARC 
						(
							-- DELETE MOUSE CALLBACK ARC
							if MCall do ( MCall = false ; delete MArcID )
							
							MakeArc id:PItem endpoint:P2 origin:P1 trans:Tform matrix:WorkTM memory:TypeRecord
							append TypeRecord #ARC
							format Prompts[LPrompt = 3]
						)
						-- SAVE POINT FOR USE IN PICKPOINT RUBBERBAND OPTION
						pt = R[1]
					)
					
				)
				
			2: 	( -- CHANGE TO ARC DRAWING MODE
					DrawMode = #ARC	
					if MCall do ( MCall = false ; delete MArcID )
					format Prompts[LPrompt = 3]
					if ((numKnots PItem 1) == 0 ) do (StartWithArc = true)
				)
 
			3: 	( -- CHANGE TO LINE DRAWING MODE
					DrawMode = #LINE 
					if MCall do (MCall = false ; delete MArcID)
					format Prompts[LPrompt = 2]
				)
				
			4: 	( -- CLOSE PLINE AND EXIT
					if DrawMode == #LINE then
					(
						Close PItem 1
						updateshape PItem
					)
					else
					(
						if MCall do delete MArcID 
						MakeArc id:PItem endpoint:P1 origin:P1 trans:TForm matrix:WorkTM memory:TypeRecord
						nK = numKnots PItem 1
						TempVec = getInVec PItem 1 nK
						deleteKnot PItem 1 nK
						setKnotType PItem 1 1 #beziercorner
						close PItem 1
						setInVec PItem 1 1 TempVec
						updateshape PItem
					)
					CarryOn = false
				)
				
			5: 	( -- UNDO SELECTED
					if MCall do ( MCall = false ; delete MArcID )
			
					-- If there are any items left to undo proceed
					if (TypeRecord.count > 0) do
					(
						Type = TypeRecord[TypeRecord.count]
						case Type of
						(
							#LINE:	UndoRtnVal = UndoLine id:PItem trans:TForm matrix:WorkTM 
							#ARC:	UndoRtnVal = UndoArc id:PItem trans:TForm matrix:WorkTM memory:TypeRecord
						)
						
						-- Set DrawMode variable to either line or arc
						DrawMode = Type
						
						-- Remove item just deleted from typeRecord array
						deleteItem TypeRecord (TypeRecord.count)	
						
						-- If there was only one entity left undo would delete it and create new
						-- spline (PItem). This returned name is either the original or the new.	
						PItem = UndoRtnVal[1]
						
						-- This is the new rubberband point, simply the last point in spline (PITEM)
						Pt = UndoRtnVal[2]	
						
						-- If the undo function deletes the PItem it returns Pt=undefined, in that case
						-- we set it back to original point selected.
						if (Pt == undefined) do Pt=P1
						
						-- Starting entity is Arc, set variable for special handling in mouse move call back
						if (TypeRecord.count == 0 and DrawMode == #ARC) do (StartWithArc = true)
						
						-- Send appropriate prompt to Listener				
						if DrawMode == #LINE 
						then format Prompts[LPrompt = 2]
						else format Prompts[LPrompt = 3]
						redrawviews()
					)						
				)
		
			6: 	( -- ENTER KEY PRESS
					if MCall do delete MArcID
					-- Check if any points have been created, if not delete PItem
					if (numKnots PItem 1 == 0) do delete PItem
					CarryOn = false
				)
				
			8: 	(-- ESCAPE KEY PRESS
					delete PItem
					redrawviews()
					setListenerSel #(-1,-1)
					return (format "*Cancel*\nOK\n")
				)
				
			9: 	( -- RIGHT MOUSE CLICK
							
					-- Get last key entered
					CaretPos = getListenerSel()
					setListenerSel #((CaretPos[1]-1),CaretPos[2])
					SelText = getListenerSelText()
					case of
					(
						(SelText=="A" or SelText=="a"): 
						(
							DrawMode=#Arc 
							if MCall do ( MCall = false ; delete MArcID )
							format Prompts[LPrompt = 3]
							if ((numKnots PItem 1) == 0 ) do (StartWithArc = true)
						)
						(SelText=="L" or SelText=="l"): 
						(
							DrawMode=#Line 
							if MCall do (MCall = false ; delete MArcID)
							format Prompts[LPrompt = 2]
						)
						(SelText=="U" or SelText=="u"):
						(
							if MCall do ( MCall = false ; delete MArcID )
						
							-- If there are any items left to undo proceed
							if (TypeRecord.count > 0) do
							(
								Type = TypeRecord[TypeRecord.count]
								case Type of
								(
									#LINE:	UndoRtnVal = UndoLine id:PItem trans:TForm matrix:WorkTM 
									#ARC:	UndoRtnVal = UndoArc id:PItem trans:TForm matrix:WorkTM memory:TypeRecord
								)
								
								-- Set DrawMode variable to either line or arc
								DrawMode = Type
								
								-- Remove item just deleted from typeRecord array
								deleteItem TypeRecord (TypeRecord.count)	
								
								-- If there was only one entity left undo would delete it and create new
								-- spline (PItem). This returned name is either the original or the new.	
								PItem = UndoRtnVal[1]
								
								-- This is the new rubberband point, simply the last point in spline (PITEM)
								Pt = UndoRtnVal[2]	
								
								-- If the undo function deletes the PItem it returns Pt=undefined, in that case
								-- we set it back to original point selected.
								if (Pt == undefined) do Pt=P1
								
								-- Starting entity is Arc, set variable for special handling in mouse move call back
								if (TypeRecord.count == 0 and DrawMode == #ARC) do StartWithArc = true
								
								-- Send appropriate prompt to Listener				
								if DrawMode == #LINE 
								then format Prompts[LPrompt = 2]
								else format Prompts[LPrompt = 3]
								redrawviews()
							)	
						)						
						(SelText=="C" or SelText=="c"):
						(
							if DrawMode == #LINE then
							(
								Close PItem 1
								updateshape PItem
							)
							else
							(
								if MCall do delete MArcID 
								MakeArc id:PItem endpoint:P1 origin:P1 trans:TForm matrix:WorkTM memory:TypeRecord
								nK = numKnots PItem 1
								TempVec = getInVec PItem 1 nK
								deleteKnot PItem 1 nK
								setKnotType PItem 1 1 #beziercorner
								close PItem 1
								setInVec PItem 1 1 TempVec
								updateshape PItem
							)
							CarryOn = false
						)
						default: 
						(
							if MCall do delete MArcID
							
							-- Check if any points have been created, if not delete PItem
							if (numKnots PItem 1 == 0) do delete PItem
							CarryOn = false
							format "\n"
						)
					)
				)
				
			default: (CarryOn = false)
		)
	)
	max views redraw
	EscapeEnable = true
	format "OK\n"
)